' =============================
' The Labyrinth of the Minotaur
' =============================
' The original QBASIC version
' (C) Copyright 1996, 1999 David E. Down, the Faery Monarch
'
' Released under the GNU GPL version 2
' Please see "COPYING" for software license details
'
DECLARE SUB PLAYLAB ()
DECLARE SUB ASKLAB ()
DECLARE SUB STALL ()
DECLARE SUB INR ()
DECLARE SUB PAUSE (P!)
DECLARE SUB CENRE (Y%, C%, T$)
DECLARE SUB WIN (X%, Y%, L%, D%, B%)
DECLARE SUB SHWLOC (X%, Y%, V%)
DECLARE SUB UNVEIL ()
DECLARE SUB CREATE ()
DECLARE FUNCTION TESTCOM% (X%, Y%)
DECLARE FUNCTION OUTMAP% (X%, Y%, DIR%)
DECLARE SUB SHWMAP (X%, Y%, DIR%)
DECLARE SUB STPRO ()
DEFINT A-Z

' The labyrinth of the Minotaur
' =============================
'  Written by:   David E. Down
'  Version:      1.0
'  Completed by: 06/JUL/1996

CONST SWG = 640' Screen Width [Graphics Pixel] (640)
CONST SDG = 350' Screen Depth [Graphics Pixel] (350)
CONST SWT = 80' Screen Width [Text Characters] (80)
CONST CYS = 14' Character Y size (14)

DIM SHARED MAP(79, 41, 1), DIR(3, 1), NULC, HDEMAP, WW, DD, SPX, SPY, MVE
DIM SHARED LL

STPRO' --- Setup program
INR' --- Introduction

 DO
  ASKLAB' --- Ask about Labyrinth Type
  HDEMAP = -1
  CLS
  CALL CENRE(11, 1, "Creating the labyrinth")
  CREATE
  UNVEIL
  PLAYLAB
  ' --- The labyrinth has been finished
  CALL CENRE(1, 14, "You have escaped the Labyrinth of the Minotaur")
  CALL CENRE(11, 1, "You took" + STR$(MVE) + " moves!")
  CALL CENRE(15, 1, "Do you want to play again? (Y/N)")
   DO
    a$ = UCASE$(INKEY$)
   LOOP UNTIL a$ = "Y" OR a$ = "N"
 LOOP UNTIL a$ = "N"

' --- End program
SCREEN 0, 1, 0: CLS : END

' --- Direction Data for the algorithm
DIR:
 DATA 0,-1,1,0,0,1,-1,0


' NOTES ON GLOBAL VARIABLES:
' ==========================
' CYS:    This is the Y-length of the text, this should be set on 14 because
'         of the graphics mode used in this program. [See main moudle, ASKLAB
'         & WIN]
' DD:     Y-Size (Depth) of labyrinth. [See procedures: ASKLAB, CREATE,
'         OUTMAP, PLAYLAB & UNVEIL]
' HDEMAP: To see how the labyrinth been constructed. Go to the main moudle of
'         the program, place ' before HDEMAP = -1 & CALL CENTRE (11, 1,
'         "Creating the labyrinth") & place a END statement between CREATE &
'         UNVEIL. [See main moudle & procedures SHWMAP & UNVEIL]
' LL:     The higher setting of this variable will cause the labyrinth to be
'         simpler to complete & lower numbers will cause the labyrinth to
'         contain lots of twist & bends plus it might make the labyrinth
'         longer to complete. [See procedure: ASKLAB, CREATE]
' MVE:    Number of moves that the player took so far. [See main moudle &
'         procedure PLAYLAB]
' NULC:   The remaining unused components of the labyrinth. [See procedures:
'         CREATE & SHWMAP]
' SDG:    The Y-Size (Depth) of the screen in pixels, this should be set on
'         350 because of the graphics mode used in this program. [See main
'         moudle & procedure ASKLAB]
' SWG:    The X-Size (Width) of the screen in pixels, this should be set on
'         640 because of the graphics mode used in this program. [See main
'         moudle & procedure ASKLAB]
' SWT:    This the number of characters can fit X-ways across the screen,
'         this should be set on 80 because of graphics mode used in this
'         program. [See main moudle & Procedure CENRE]
' SPX:    X-size used to show on screen components [See procedures: ASKLAB,
'         SHWLOC & SHWMAP]
' SPY:    Y-size used to show on screen components [See procedures: ASKLAB,
'         SHWLOC & SHWMAP]
' WW:     X-Size (Width) of labyrinth. [See procedures: ASKLAB, CREATE,
'         OUTMAP, PLAYLAB & UNVEIL]

' The algorithm that creates the labyrinth contained in the procedures
' CREATE, OUTMAP, SHWMAP & TESTCOM. This algorithm uses global variables
' WW, DD, NULC, SPX, SPY, HDEMAP & arrays MAP & DIR. The array DIR been set
' up in the procedure STPRO

' Guess what? The labyrinths that the algorithm creates, they will only have
' one way out. (Ha! Ha! Ha!)

' The procedure that the player gets to choose its labyrinth type is in the
' procedure ASKLAB, Player gets to explore the labyrinth in the procedure
' PLAYLAB.

' --- Ask for the Labyrinth Difficulty
SUB ASKLAB
 CLS
 CALL CENRE(1, 1, "Welcome to the Labyrinth of the Minotaur")
 CALL WIN(31, 5, 17, 10, 14)
 LOCATE 6, 32: PRINT "Select Difficulty"
 LOCATE 8, 32: PRINT "1. Petty Easy"
 LOCATE 9, 32: PRINT "2. Easy"
 LOCATE 10, 32: PRINT "3. Near Average"
 LOCATE 11, 32: PRINT "4. About Average"
 LOCATE 12, 32: PRINT "5. Above Average"
 LOCATE 13, 32: PRINT "6. Hard"
 LOCATE 14, 32: PRINT "7. Very Hard"
 LOCATE 15, 32: PRINT "8. Mega Hard"
 CALL CENRE(18, 1, "Please press the number of your desired choice")
  DO
   a$ = INKEY$
  LOOP UNTIL a$ > "0" AND a$ < "9"
 ' --- Select Labyrinth Size Variables
  SELECT CASE a$
   CASE "1"' --- Selection 1
    WW = 16: DD = 16: LL = 2
   CASE "2"' --- Selection 2
    WW = 32: DD = 24: LL = 4
   CASE "3"' --- Selection 3
    WW = 40: DD = 24: LL = 5
   CASE "4"' --- Selection 4
    WW = 40: DD = 32: LL = 5
   CASE "5"' --- Selection 5
    WW = 52: DD = 32: LL = 6
   CASE "6"' --- Selection 6
    WW = 64: DD = 33: LL = 8
   CASE "7"' --- Selection 7
    WW = 80: DD = 41: LL = 10
   CASE "8"' --- Selection 8
    WW = 80: DD = 41: LL = 5
  END SELECT
' --- Set final
 SPX = SWG \ WW
 SPY = (SDG - CYS) \ DD

END SUB

' --- Place a centred a boxed message
SUB CENRE (Y, C, T$)
 PS = (SWT - LEN(T$)) \ 2 + 1
 CALL WIN(PS - 1, Y, LEN(T$), 1, C)
 LOCATE Y + 1, PS: PRINT T$;
END SUB

' --- Create the labyrinth
SUB CREATE
EN = -1' --- Set singal to first create entrance
NULC = WW * DD' --- Find Labyrinth Area

 DO
   IF EN THEN ' --- Start with labyrinth entrance
     X = 0
     Y = INT(RND * DD)
     MAP(X, Y, 1) = 1
     MAP(X, Y, 0) = 8: NULC = NULC - 1
     DIR = TESTCOM(X, Y)
     EN = 0
    ELSE ' --- Start next labyrinth segment
     X = INT(RND * WW): Y = INT(RND * DD): OX = X: OY = Y
      ' --- Find the next available start off component
      DO
       DIR = TESTCOM(X, Y)
       IF DIR < 0 THEN
	X = X + 1: IF X = WW THEN X = 0: Y = Y + 1: IF Y = DD THEN Y = 0
       ELSE
	EXIT DO
       END IF
      LOOP
   END IF

  ' --- Draw Labyrinth segment
   DO
    LN = INT(RND * LL + 1)
     DO
      OX = X: OY = Y: PL = MAP(X, Y, 1) + 1
      ' --- Check if in boundary or not connecting another labyrinth segment
      IF OUTMAP(X, Y, DIR) THEN EXIT DO
      ' --- Draw labyrinth component
      X = X + DIR(DIR, 0): Y = Y + DIR(DIR, 1)
      CALL SHWMAP(OX, OY, 2 ^ DIR)
      MAP(X, Y, 1) = PL
      CALL SHWMAP(X, Y, 2 ^ (DIR XOR 2)): LN = LN - 1
     LOOP UNTIL LN = 0
    ' --- Check if able to contiune labyrinth segment
    ORG = DIR: DIR = (DIR + 1 + INT(RND * 2) * 2) MOD 4
    FLG = -1: IF NOT OUTMAP(X, Y, DIR) THEN FLG = 0
    IF FLG THEN DIR = DIR XOR 2: IF NOT OUTMAP(X, Y, DIR) THEN FLG = 0
    IF FLG THEN DIR = ORG: IF NOT OUTMAP(X, Y, DIR) THEN FLG = 0
   LOOP UNTIL FLG
 LOOP UNTIL NULC = 0

 ' --- Place Labyrinth Exit (at a place with the most number of links)
 HH = 0: X$ = ""
  FOR Y = 0 TO DD - 1
   NOL = MAP(WW - 1, Y, 1)
   IF NOL > HH THEN HH = NOL: X$ = ""
   IF NOL = HH THEN X$ = X$ + CHR$(Y)
  NEXT
 LC = ASC(MID$(X$, INT(RND * LEN(X$)) + 1))
 CALL SHWMAP(WW - 1, LC, 2)

END SUB

' --- The introduction procedure
SUB INR
 CLS
 CALL CENRE(1, 1, "Welcome to the Labyrinth of the Minotaur")
 CALL WIN(5, 5, 68, 5, 14)
  FOR I = 6 TO 10
   LOCATE I, 6: FL = -1
    SELECT CASE I
     CASE 6
      PRINT "How dare you enter Labyrinth of the Minotaur?"
     CASE 7
      FL = 0
     CASE 8
      PRINT "There will be no turning back! The entrance is now been sealed off,"
     CASE 9
      PRINT "Your fate of escaping the labyrinth & returning back to your own"
     CASE 10
      PRINT "world relies on the exit on the other side of the labyrinth!"
    END SELECT
   IF FL THEN PAUSE (2)
  NEXT
 STALL
END SUB

' --- Checks if out of boundary or may connects another labyrinth segment
FUNCTION OUTMAP (X, Y, DIR)
 XX = X + DIR(DIR, 0)
 YY = Y + DIR(DIR, 1)
 IF XX < 0 OR YY < 0 OR XX = WW OR YY = DD THEN OUTMAP = -1: EXIT FUNCTION
 IF MAP(XX, YY, 0) > 0 THEN OUTMAP = -1 ELSE OUTMAP = 0
END FUNCTION

' --- The Pausing Procedure
SUB PAUSE (P!)
 V! = TIMER + P!
 DO: LOOP UNTIL TIMER > V!
END SUB

' --- Play labyrinth
SUB PLAYLAB
 ' --- Find Start
 X = 0: Y = 0
  FOR I = 0 TO DD - 1
   IF (MAP(0, I, 0) AND 8) THEN Y = I: I = DD
  NEXT
 ' --- Seal Entrance
 MAP(X, Y, 0) = MAP(X, Y, 0) AND 7
 LINE (X * SPX, Y * SPY + 1)-(X * SPX, (Y + 1) * SPY - 2), 5
 MVE = 0

' --- Play with labyrinth
  DO
   CALL SHWLOC(X, Y, 15)
    ' --- GET KEY
    DO
     a$ = INKEY$
    LOOP UNTIL a$ > ""
   CALL SHWLOC(X, Y, 0)
    SELECT CASE a$
     CASE CHR$(0) + "H"' --- UP
      IF MAP(X, Y, 0) AND 1 THEN Y = Y - 1: MVE = MVE + 1
     CASE CHR$(0) + "P"' --- DOWN
      IF MAP(X, Y, 0) AND 4 THEN Y = Y + 1: MVE = MVE + 1
     CASE CHR$(0) + "K"' --- LEFT
      IF MAP(X, Y, 0) AND 8 THEN X = X - 1: MVE = MVE + 1
     CASE CHR$(0) + "M"' --- RIGHT
      IF MAP(X, Y, 0) AND 2 THEN X = X + 1: MVE = MVE + 1
    END SELECT
  LOOP UNTIL X = WW

END SUB

' --- Show Location
SUB SHWLOC (X, Y, V)
 LINE (X * SPX + 2, Y * SPY + 2)-((X + 1) * SPX - 3, (Y + 1) * SPY - 3), V, BF
END SUB

' --- Draw part of labyrinth map
'
SUB SHWMAP (X, Y, DIR)
 ' --- Subtract remaining labyrinth components
 IF MAP(X, Y, 0) = 0 THEN NULC = NULC - 1
 XN = MAP(X, Y, 0) OR DIR: MAP(X, Y, 0) = XN
 IF HDEMAP THEN EXIT SUB
 ' --- Draw Wall Corners
 PSET (X * SPX, Y * SPY), 10
 PSET ((X + 1) * SPX - 1, Y * SPY), 10
 PSET ((X + 1) * SPX - 1, (Y + 1) * SPY - 1), 10
 PSET (X * SPX, (Y + 1) * SPY - 1), 10
 ' --- North Wall
 LINE (X * SPX + 1, Y * SPY)-((X + 1) * SPX - 2, Y * SPY), ((XN AND 1) = 0) * -10
 ' --- East Wall
 LINE ((X + 1) * SPX - 1, Y * SPY + 1)-((X + 1) * SPX - 1, (Y + 1) * SPY - 2), ((XN AND 2) = 0) * -10
 ' --- South Wall
 LINE (X * SPX + 1, (Y + 1) * SPY - 1)-((X + 1) * SPX - 2, (Y + 1) * SPY - 1), ((XN AND 4) = 0) * -10
 ' --- West Wall
 LINE (X * SPX, Y * SPY + 1)-(X * SPX, (Y + 1) * SPY - 2), ((XN AND 8) = 0) * -10
 ' --- Record New Wall Placement
 MAP(X, Y, 0) = XN
END SUB

' --- Ask the user to press the SPACE-BAR
SUB STALL
 CALL CENRE(23, 1, "Please press the SPACE-BAR to contiune")
 DO: LOOP UNTIL INKEY$ = " "
END SUB

' --- Set up the labyrinth
SUB STPRO
 SCREEN 9
 ' --- Randomize the random generator
 RANDOMIZE TIMER
 ' --- Get Direction Data
 RESTORE DIR
  FOR I = 0 TO 3
   READ DIR(I, 0), DIR(I, 1)
  NEXT I

END SUB

' --- Test for blank adjacent labyrinth components
'
' TESTCOM parameters:
' ===================
'  X: X-axis of component
'  Y: Y-axis of component
'
' TESTCOM = Returns a random direction
FUNCTION TESTCOM (X, Y)
 IF MAP(X, Y, 0) = 0 THEN TESTCOM = -1: EXIT FUNCTION
 FLG = 0
  FOR I = 0 TO 3
   IF OUTMAP(X, Y, DIR) THEN FLG = -1 ELSE DIR = (DIR + 1) MOD 4
  NEXT
 IF FLG THEN TESTCOM = DIR ELSE TESTCOM = -1
END FUNCTION

' --- Show the labyrinth on screen
SUB UNVEIL

 CLS
 HDEMAP = 0
  FOR X = 0 TO WW - 1
   FOR Y = 0 TO DD - 1
    CALL SHWMAP(X, Y, 0)
   NEXT
  NEXT

END SUB

' --- Draw a window
'
' WIN parameters
' ==============
'  X: X-position of window
'  Y: Y-position of window
'  L: Length of Window
'  D: Depth of Window
'  B: Border Colour of Window
SUB WIN (X, Y, L, D, B)
 XX = (X - 1) * 8: YY = (Y - 1) * CYS: LS = L * 8: DS = D * CYS
 LINE (XX, YY)-(XX + LS + 15, YY + DS + CYS * 2 - 1), 0, BF
 LINE (XX + 3, YY + CYS \ 2 - 1)-(XX + LS + 12, YY + DS + CYS + CYS \ 2), B, B
 LINE (XX + 4, YY + CYS \ 2)-(XX + LS + 11, YY + DS + CYS + CYS \ 2 - 1), B, B
END SUB

