/* =============================
   The Labyrinth of the Minotaur
   =============================
   Written by:   David E. Down
   Version:      1.0
-- QBASIC version --
   Completed on: 06/JUL/1996
-- DJGPP/ALLEGRO conversion --
   Completed on: 06/AUG/1999

   Copyright David E. Down, the Faery Monarch 1996, 1999

   Released under the GNU GPL version 2.
   Please see "COPYING" for software license details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <bios.h>
#include <string.h>
#include <allegro.h>
#include "def.h"

FONT *biosfont;
CHAR biosfontheight;

/* -- The Colour Palette -- */
RGB clrs[] = {{0x00, 0x00, 0x00}, {0x00, 0x00, 0x20},
              {0x00, 0x20, 0x00}, {0x00, 0x20, 0x20},
              {0x20, 0x00, 0x00}, {0x20, 0x00, 0x20},
              {0x20, 0x20, 0x00}, {0x2a, 0x2a, 0x2a},
              {0x15, 0x15, 0x15}, {0x00, 0x00, 0x3f},
              {0x00, 0x3f, 0x00}, {0x00, 0x3f, 0x3f},
              {0x3f, 0x00, 0x00}, {0x3f, 0x00, 0x3f},
              {0x3f, 0x3f, 0x00}, {0x3f, 0x3f, 0x3f}};

/* -- The Procedures -- */
void playlab ();
void asklab ();
void stall ();
void inr ();
void pause (float p);
void cenre (int y, int c, char *t);
void win (int x, int y, int l, int d, int b);
void shwloc (int x, int y, int v);
void unveil ();
void create ();
int testcom (int x, int y);
bool outmap (int x, int y, int dir);
void shwmap (int x, int y, int dir);
void stpro ();

int swg = 640; /* Screen Width [Graphics Pixel] (640) */
int sdg = 400; /* Screen Depth [Graphics Pixel] (480) */
int swt = 80;  /* Screen Width [Text Characters] (80) */
int cys = 16;   /* Character Y size (14) */

unsigned char map[80][42][2];
char dirs[4][3] = {{0, -1, 1}, {1, 0, 2}, {0, 1, 4}, {-1, 0, 8}};
bool hdemap;
int nulc, ww, dd, spx, spy, mve, ll;

void pause (float p) {
  int i = biostime(0, 0) + (int)(p * 18.2); /* It is safe to use this */

   do {
#ifdef DEBUG_ESC
     if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
   } while (i > biostime (0, 0));
 }

/* Show Location */
void shwloc (x, y, v) {
  rectfill(screen, x * spx + 2, y * spy + 2, (x + 1) * spx - 3, (y + 1) * spy - 3, v);
 }


/* Checks if out of boundary or may connect another labyrinth segment */
bool outmap (x, y, dir) {
  int xx = x + dirs[dir][0], yy = y + dirs[dir][1];

  if (xx < 0 || yy < 0 || xx >= ww || yy >= dd) { return TRUE; }
  if (map[xx][yy][0] > 0) { return TRUE; }
  return FALSE;
 }

/*
 * --- Test for blank adjacent labyrinth components
 *
 * TESTCOM parameters:
 * ===================
 *  X: X-axis of component
 *  Y: Y-axis of component
 *
 * TESTCOM = Returns a random direction
 */
int testcom(int x, int y) {
 bool flg = FALSE;
 char i, dir = random () & 0x3;

 if (map[x][y][0] == 0) { return -1; }
  for (i = 0; i < 4; i++) {
   if (outmap(x, y, dir)) { flg = TRUE; } else { dir = (dir + 1) & 0x3; }
  }
 if (flg) { return dir; } else { return -1; }
}

/* --- Draw part of labyrinth map --- */
void shwmap (x, y, dir) {
 int xn;

 /* --- Subtract remaining labyrinth components --- */
  if (map[x][y][0] == 0) nulc--;
  xn = map[x][y][0] | dir; map[x][y][0] = xn;
#ifndef DEBUG_TESTMAP
  if (hdemap) return;
#endif
 /* --- Draw Wall Corners --- */
  putpixel(screen, x * spx, y * spy, 10);
  putpixel(screen, (x + 1) * spx - 1, y * spy, 10);
  putpixel(screen, (x + 1) * spx - 1, (y + 1) * spy - 1, 10);
  putpixel(screen, x * spx, (y + 1) * spy - 1, 10);

 /* -- North Wall -- */
  hline(screen, x * spx + 1, y * spy, (x + 1) * spx - 2, (xn & 1 ? 0 : 10));
 /* -- East Wall -- */
  vline(screen, (x + 1) * spx - 1, y * spy + 1, (y + 1) * spy - 2, (xn & 2 ? 0 : 10));
 /* -- South Wall -- */
  hline(screen, x * spx + 1, (y + 1) * spy - 1, (x + 1) * spx - 2, (xn & 4 ? 0 : 10));
 /* -- West Wall -- */
  vline(screen, x * spx, y * spy + 1, (y + 1) * spy - 2, (xn & 8 ? 0 : 10));

 }

/* Create the labyrinth */
void create () {
 bool en = TRUE, flg = FALSE;
 int x, y, ox, oy, org, dir, pl, nol, lc, ln, hh = 0;

 nulc = ww * dd;
 memset (map, 0x00, sizeof(map)); /* Clear the map array to zero */
  do {
   if (en) {
     x = 0;
     y = random () % dd;
     map[x][y][1] = 1;
     map[x][y][0] = 8; nulc--;
     dir = testcom(x,y);
     en = FALSE;
   } else {
 /* --- Find the next available start off component --- */
    x = random () % ww; y = random () % dd;
#ifdef DEBUG
    fprintf(stderr, "labyrin.c: create () is finding next start off component\n");
#endif
     for (;;) {
#ifdef DEBUG_ESC
      if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
      dir = testcom(x, y);
      if (dir < 0) {
       x++; if (x >= ww) { x = 0; y++; if (y >= dd) y = 0; }
      } else {
       break;
      }
     }
   }
/* --- Draw Labyrinth segment --- */
#ifdef DEBUG
  fprintf(stderr, "labyrin.c: create () is drawing a labyrinth segment\n");
#endif
   do {
    ln = (random () % ll) + 1;
     do {
#ifdef DEBUG_ESC
      if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
      ox = x; oy = y; pl = map[x][y][1] + 1;
      /* -- Check if in boundary or not connecting another labyrinth segment -- */
      if (outmap(x, y, dir)) break;
      /* -- Draw labyrinth component -- */
      x += dirs[dir][0]; y += dirs[dir][1];
      shwmap(ox, oy, dirs[dir][2]); map[x][y][1] = pl;
      shwmap(x, y, dirs[0x2 ^ dir][2]); ln--;
     } while (!ln);
    /* -- Check if able to contiune labyrinth segment -- */
#ifdef DEBUG
    fprintf(stderr, "labyrin.c: create () is checking to see the labyrinth segment can continue\n");
#endif
#ifdef DEBUG_ESC
    if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
    org = dir; dir = (dir + 1 + (random () & 0x2)) & 0x3;
    flg = (outmap(x, y, dir) ? TRUE : FALSE);
    if (flg) {dir ^= 2; if (!outmap(x, y, dir)) flg = FALSE; }
    if (flg) {dir = org; if (!outmap(x, y, dir)) flg = FALSE; }
   } while (!flg);
  } while (nulc);

/* --- Place Labyrinth Exit (at a place with the most number of links) --- */
#ifdef DEBUG
    fprintf(stderr, "labyrin.c: create () is creating a labyrinth exit\n");
#endif

  for (y = 0; y < dd; y++) {
    nol = map[ww - 1][y][1];
    if (nol > hh) { hh = nol; lc = y; }
   }
  shwmap(ww - 1, lc, 2);
}

/* Setup the program */
void stpro () {
  srandom(biostime(0,0));
  allegro_init ();
  set_color_depth (8);
   if (set_gfx_mode (GFX_AUTODETECT, swg, sdg, swg, sdg)) {
    printf("Error! Unable to setup graphics.\n"); exit(255);
   }
  set_palette_range (clrs, 0, 15, TRUE);
  biosfont = font; biosfontheight = text_height (biosfont);

  install_keyboard ();
 }

/* Just to assist in porting this program to C */
static inline void loc_print (int y, int x, char *t) {
 #ifdef USE_BIOSFONT
  textout(screen, biosfont, t, (x - 1) << 3,
          ((y - 1) << 4) + ((cys - biosfontheight) >> 1), 15);
 #else
 /* Currently, there is only the BIOS font to use at this moment */
  textout(screen, biosfont, t, (x - 1) << 3,
          ((y - 1) << 4) + ((cys - biosfontheight) >> 1), 15);
 #endif
 }

/* Draw a window on the screen
 *
 * X: X position of window
 * Y: Y position of window
 * L: Length of window
 * D: Depth of window
 * B: Border color of window
 */
void win (int x, int y, int l, int d, int b) {
  int xx = (x - 1) << 3;
  int yy = (y - 1) * cys;
  int ls = l << 3;
  int ds = d * cys;

  rectfill(screen, xx, yy, xx + ls + 15, yy + ds + (cys << 1), 0);
  rect(screen, xx + 3, yy + (cys >> 1) - 1,
               xx + ls + 12, yy + ds + cys + (cys >> 1), b);
  rect(screen, xx + 4, yy + (cys >> 1),
               xx + ls + 11, yy + ds + cys + (cys >> 1) - 1, b);
 }

/* Place a message at a centre of screen line */
 void cenre (int y, int c, char *t) {
   int ps = ((swt - strlen(t)) >> 1) + 1;

   win (ps - 1, y, strlen(t), 1, c);
   loc_print(y + 1, ps, t); /* Assist porting to C */
  }

 /* The Stall Function */
void stall () {
 bool keyflag = TRUE;

  cenre(23, 1, "Please press the SPACE-BAR to contiune");
   do {
#ifdef DEBUG_ESC
    if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
     if (keypressed ()) {
      if ((readkey () & 0xff) == 0x20) keyflag = FALSE;
     }
   } while (keyflag);

 }

 /* The Introduction */
void inr () {
 int i; char fl;

 clear(screen);
 cenre(1, 1, "Welcome to the Labyrinth of the Minotaur");
 win(5, 5, 68, 5, 14);

  for (i = 6; i < 11; i++) {
   fl = TRUE;
    switch (i) {
     case 6: {
      loc_print (i, 6, "How dare you enter Labyrinth of the Minotaur?");
      break; }
     case 7: {
      fl = FALSE;
      break; }
     case 8: {
      loc_print (i, 6, "There will be no turning back! The entrance is now been sealed off,");
      break; }
     case 9: {
      loc_print (i, 6, "Your fate of escaping the labyrinth & returning back to your own");
      break; }
     case 10: {
      loc_print (i, 6, "world relies on the exit on the other side of the labyrinth!");
      break; }
    }
   if (fl) pause(2);
  }
 stall ();
 }

/* Ask for the Labyrinth Difficulty */
void asklab () {
 CHAR a = 0;
 int aa;

 clear(screen);
 cenre(1, 1, "Welcome to the Labyrinth of the Minotaur");
 win(31, 5, 17, 10, 14);
 loc_print(6, 32, "Select Difficulty");
 loc_print(8, 32, "1. Petty Easy");
 loc_print(9, 32, "2. Easy");
 loc_print(10,32, "3. Near Average");
 loc_print(11, 32, "4. About Average");
 loc_print(12, 32, "5. Above Average");
 loc_print(13, 32, "6. Hard");
 loc_print(14, 32, "7. Very Hard");
 loc_print(15, 32, "8. Mega Hard");
 cenre(18, 1, "Please press the number of your desired choice");
  do {
   if (keypressed ()) {
    aa = readkey ();
#ifdef DEBUG_ESC
    if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
    a = aa;
   }
  } while (a < '1' || a > '8');

 /* Select Labyrinth Size Variables */
  switch (a) {
    case '1': { ww = dd = 16; ll = 2; break; }
    case '2': { ww = 32; dd = 24; ll = 4; break; }
    case '3': { ww = 40; dd = 24; ll = 5; break; }
    case '4': { ww = 40; dd = 32; ll = 5; break; }
    case '5': { ww = 52; dd = 32; ll = 6; break; }
    case '6': { ww = 64; dd = 33; ll = 8; break; }
    case '7': { ww = 80; dd = 41; ll = 10; break; }
    case '8': { ww = 80; dd = 41; ll = 5; break; }
    default: {
    #ifdef DEBUG
     fprintf(stderr, "labyrin.c: switch could not find a case\n");
    #endif
     break; }
   }
 /* Set Final */
#ifdef DEBUG
 fprintf(stderr, "labyrin.c: ww = %U; dd = %u; ll = %u;\n", ww, dd, ll);
#endif
  spx = (swg / (int) ww);
  spy = (sdg - cys) / (int) dd;

}

/* -- Show the labyrinth on screen -- */
void unveil () {
 int x, y;

 clear (screen);
 hdemap = FALSE;
  for (x = 0; x < ww; x++) {
   for (y = 0; y < dd; y++) {
    shwmap(x,y,0);
   }
  }

}

/* -- Play labyrinth -- */
void playlab () {
 CHAR a;
 int aa, x, y, i;

 /* --- Find Start */
 x = y = 0;
  for (i = 0; i < dd; i++) {
   if (map[0][i][0] & 0x8) { y = i; break; }
  }

 /* -- Seal Entrance -- */
 map[x][y][0] &= 0x7;
 vline(screen, x * spx, y * spy + 1, (y + 1) * spy - 2, 5);
 mve = 0;


 /* -- Play with labyrinth -- */
  do {
   shwloc(x, y, 15);
   a = 0x00;
    do {
#ifdef DEBUG_ESC
     if (key[KEY_ESC]) exit(255); /* Ye PANIC key */
#endif
      if (keypressed ()) {
       aa = readkey ();
       if (!(aa & 0xff)) a = aa >> 8;
      }
    } while (!a);
   shwloc (x, y, 0); /* -- Remove Cursor -- */
    switch (a) {
     case 'H': { /* -- UP -- */
      if (map[x][y][0] & 0x1) { y--; mve++; }
      break; }
     case 'P': { /* -- DOWN -- */
      if (map[x][y][0] & 0x4) { y++; mve++; }
      break; }
     case 'K': { /* -- LEFT -- */
      if (map[x][y][0] & 0x8) { x--; mve++; }
      break; }
     case 'M': { /* -- RIGHT -- */
      if (map[x][y][0] & 0x2) { x++; mve++; }
      break; }
    }
  } while (x != ww);
 }

/* The main procedure */
int main () {
 CHAR text[81], a;

 stpro ();
 inr ();


 do {
  asklab (); /* Ask about Labyrinth Type */
  hdemap = TRUE;
  clear (screen);
  cenre(11, 1, "Creating the labyrinth");
  create ();
  unveil ();
  playlab ();
/* -- The labyrinth has been finished -- */
  cenre(1, 14, "You have escaped the Labyrinth of the Minotaur");
  sprintf(text, "You took %D moves!", mve); cenre(11, 1, text);
  cenre(15, 1, "Do you want to play again? (Y/N)");

  a = 0x0;
   do {
    if (keypressed ()) {
     a = readkey ();
    }
   } while (!(a == 'y' || a == 'Y' || a == 'n' || a == 'N'));
 } while (!(a == 'n' || a == 'N'));

/* -- End program -- */
  set_gfx_mode (GFX_TEXT, 0, 0, 0, 0);
  return 0;
 }

